/*-*- c++ -*-********************************************************
 * wxComposeView.h: a window displaying a mail message              *
 *                                                                  *
 * (C) 1998-1999 by Karsten Ballder (kasrten@phy.hw.ac.uk)         *
 *                                                                  *
 * $Id: wxComposeView.h,v 1.57 2000/09/30 16:07:27 ballueder Exp $
 *******************************************************************/

#ifndef WXCOMPOSEVIEW_H
#define WXCOMPOSEVIEW_H

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

#ifdef __GNUG__
#pragma interface "wxComposeView.h"
#endif

#ifndef USE_PCH
#   include   "Message.h"
#   include   "gui/wxMenuDefs.h"
#   include   "gui/wxMFrame.h"
#   include   "gui/wxIconManager.h"
#   include   "Profile.h"
#   include   "kbList.h"
#endif

// ----------------------------------------------------------------------------
// forward declarations
// ----------------------------------------------------------------------------
class wxComposeView;
class wxLayoutWindow;

class wxProcess;
class wxProcessEvent;

// ----------------------------------------------------------------------------
// constants
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// types and classes
// ----------------------------------------------------------------------------

struct wxCVFileMapEntry
{
   unsigned long   id;
   String          filename;
};

KBLIST_DEFINE(wxCVFileMapType,wxCVFileMapEntry);

/** A wxWindows ComposeView class
  */

class wxComposeView : public wxMFrame //FIXME: public ComposeViewBase
{
public:
   // constants

   /// header fields we may show in the composer (not all of them are shown!)
   enum AddressField
   {
      Field_From,
      Field_To,
      Field_Subject,
      Field_Cc,
      Field_Bcc,
      Field_Max
   };

   /// Do we want to send mail or post a news article?
   enum Mode
   {
      Mode_Mail,
      Mode_News
   };

   // a message kind - combined with the mode this determines the template
   // InitText() will use
   enum MessageKind
   {
      Message_New,      // either new mail message or new article
      Message_Reply,    // reply or follow-up
      Message_Forward   // this one is incompatible with Mode_NNTP
   };

   /** Constructor for posting news.
       @param parentProfile parent profile
       @param hide if true, do not show frame
       @return pointer to the new compose view
    */
   static wxComposeView * CreateNewArticle(const MailFolder::Params& params,
                                           Profile *parentProfile = NULL,
                                           bool hide = false);

   /** Constructor for sending mail.
       @param parentProfile parent profile
       @param hide if true, do not show frame
       @return pointer to the new compose view
    */
   static wxComposeView * CreateNewMessage(const MailFolder::Params& params,
                                           Profile *parentProfile = NULL,
                                           bool hide = false);

   /** Constructor for sending a reply to a message.

       @param parentProfile parent profile
       @param original message that we replied to
       @param hide if true, do not show frame
       @return pointer to the new compose view
    */
   static wxComposeView * CreateReplyMessage(const MailFolder::Params& params,
                                             Profile *parentProfile,
                                             Message * original = NULL,
                                             bool hide = false);

   /** Constructor for forwarding a message.

       @param templ is the template to use
       @param parentProfile parent profile
       @param hide if true, do not show frame
       @return pointer to the new compose view
    */
   static wxComposeView * CreateFwdMessage(const MailFolder::Params& params,
                                           Profile *parentProfile,
                                           bool hide = false);

   /** Initializes the composer text: for example, if this is a reply, inserts
       the quoted contents of the message being replied to (except that, in
       fact, it may do whatever the user configured it to do using templates).
       The msg parameter may be NULL only for the new messages, messages
       created with CreateReply/FwdMessage require it to be !NULL.

       @param msg the message we're replying to or forwarding
    */
   void InitText(Message *msg = NULL);

   /// Destructor
   ~wxComposeView();

   /// is it ok to close now?
   virtual bool CanClose() const;

   /** insert a file into buffer
       @param filename file to insert
       @param mimetype mimetype to use
       @param num_mimetype numeric mimetype
       */
   void InsertFile(const char *filename = NULL,
                   const char *mimetype = NULL);

   /** Insert MIME content data
       @param data pointer to data (we take ownership of it)
       @param len length of data
       @param mimetype mimetype to use
       @param filename optional filename to add to list of parameters
       */
   void InsertData(char *data,
                   size_t length,
                   const char *mimetype = NULL,
                   const char *filename = NULL);

   /// move the cursor to the given position
   void MoveCursorTo(int x, int y);

   /** Sets the address fields, To:, CC: and BCC:.
       @param To primary address to send mail to
       @param CC carbon copy addresses
       @param BCC blind carbon copy addresses
   */
   void SetAddresses(const String &To,
                     const String &CC = "",
                     const String & = "");

   /** Set the newsgroups to post to.
       @param groups the list of newsgroups
   */
   void SetNewsgroups(const String &groups);

   /// sets Subject field
   void SetSubject(const String &subj);

   /// get the current value of a header field
   String GetHeaderValue(AddressField which) const
   {
      wxTextCtrl *text = m_txtFields[which];
      return text ? text->GetValue() : wxString("");
   }

   /// inserts a text
   void InsertText(const String &txt);

   /// make a printout of input window
   void Print(void);

   /** Send the message.
       @param schedule if TRUE, call calendar module to schedule sending
       @return true if successful, false otherwise
   */
   bool Send(bool schedule = FALSE);

   /** wxWindows callbacks
   */
   //@{
      /// called when text zone contents changes
   void OnTextChange(wxCommandEvent &event);

      /// called on Menu selection
   void OnMenuCommand(int id);

      /// for button
   void OnExpand(wxCommandEvent &event);

      /// for identity combo
   void OnIdentChange(wxCommandEvent& event);

      /// called when external editor terminates
   void OnExtEditorTerm(wxProcessEvent& event);
   //@}

   // for wxAddressTextCtrl usage
   void SetLastAddressEntry(AddressField field) { m_fieldLast = field; }

   // for wxAddressTextCtrl usage:
   Profile *GetProfile(void) const { return m_Profile; }

   /** Adds an extra header line.
       @param entry name of header entry
       @param value value of header entry
   */
   void AddHeaderEntry(const String &entry, const String &value);

   /// reset the "dirty" flag
   void ResetDirty();

   // these functions are for backwards compatibility only
   static wxComposeView * CreateNewArticle(Profile *parentProfile = NULL,
                                           bool hide = false)
   {
      return CreateNewArticle(MailFolder::Params(""),
                              parentProfile,
                              hide);
   }

   static wxComposeView * CreateNewMessage(Profile *parentProfile = NULL,
                                           bool hide = false)
   {
      return CreateNewMessage(MailFolder::Params(""),
                              parentProfile,
                              hide);
   }

protected:
   /** quasi-Constructor
       @param parent parent window
       @param parentProfile parent profile
       @param to default value for To field
       @param cc default value for Cc field
       @param bcc default value for Bcc field
       @param hide if true, do not show frame
   */
   void Create(wxWindow *parent = NULL,
               Profile *parentProfile = NULL,
               bool hide = false);

   /** Constructor
       @param iname  name of windowclass
       @param parent parent window
   */
   wxComposeView(const String &iname, wxWindow *parent = NULL);

   // helpers
   // -------

   /// clear the window
   void DoClear();

   /// InitText() helper
   void DoInitText(Message *msg);

   /// verify that the message can be sent
   bool IsReadyToSend() const;

   /// insert a text file at the current cursor position
   bool InsertFileAsText(const String& filename,
                         bool replaceFirstTextPart = false);

   /// save the first text part of the message to the given file
   bool SaveMsgTextToFile(const String& filename,
                          bool errorIfNoText = true) const;

   /// A list of all extra headerslines to add to header.
   kbStringList m_ExtraHeaderLinesNames;
   /// A list of all extra headerslines to add to header.
   kbStringList m_ExtraHeaderLinesValues;

   /// Launch the external editor
   bool StartExternalEditor();

   /// Set the default value for the "From" header (if we have it)
   void SetFrom();

private:
   /// a profile
   Profile * m_Profile;
   /// the name of the class
   String m_name;

   /// the initial from/reply-to address
   String m_from;
   
   /// the panel
   wxPanel *m_panel;

   /// compose menu
   wxMenu *composeMenu;

   
   /// the edit/cut menu item
   class wxMenuItem *m_MItemCut;
   /// the edit/copy menu item
   class wxMenuItem *m_MItemCopy;
   /// the edit/paste menu item
   class wxMenuItem *m_MItemPaste;

   /**@name Input fields (arranged into an array) */
   //@{
      /// The text fields
   AddressField m_fieldLast;            // which had the focus last time
   wxTextCtrl  *m_txtFields[Field_Max];

      /// the canvas for displaying the mail
   wxLayoutWindow *m_LayoutWindow;
      /// the alias expand button
   wxButton *aliasButton;
   /// the button showing the xface
   class wxXFaceButton *m_XFaceButton;
   //@}

   int m_font, m_size;
   wxColour m_fg, m_bg;
   /// the popup menu
   wxMenu *popupMenu;
   /**@name The interface to its canvas. */
   //@{
   /// the ComposeView canvas class
   friend class wxCVCanvas;
   //@}

   /// a list mapping IDs to filenames
   wxCVFileMapType fileMap;
   /// for assigning IDs
   unsigned long nextFileID;
   /// find the filename for an ID
   const char *LookupFileName(unsigned long id);

   /// makes the canvas
   void CreateFTCanvas(void);

   /// enable/disable editing of the message text
   inline void EnableEditing(bool enable);

   /// ids for different processes we may launch
   enum
   {
      HelperProcess_Editor = 1
   };

   /// News or smtp?
   Mode m_mode;

   /// New article, reply/follow-up or forward?
   MessageKind m_kind;

   /// Has message been sent already?
   bool m_sent;

   /// If replying, this is the original message
   Message * m_OriginalMessage;

   /// the template to use or an empty string
   String m_template;

   /// the (main) encoding (== charset) to use for the message
   wxFontEncoding m_encoding;

   // external editor support
   // -----------------------

   wxProcess *m_procExtEdit;  // wxWin process notification helper
   String     m_tmpFileName;  // temporary file we passed to the editor
   int        m_pidEditor;    // pid of external editor (0 if none)

   DECLARE_EVENT_TABLE()
   DECLARE_CLASS(wxComposeView)
};

#endif
